# -*- coding: utf-8 -*-

# Copyright 2008-2011 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

import math
from skoolparser import SkoolParser

# The ID of the main disassembly
MAIN_CODE_ID = 'main'

class SkoolGame:
    def __init__(self, skoolfile, ref_parser, code_id=MAIN_CODE_ID):
        self.skoolfile = skoolfile
        self.ref_parser = ref_parser
        self.info = self.ref_parser.build_dictionary('Info')

        self.paths = self.ref_parser.build_dictionary('Paths')
        self.titles = self.ref_parser.build_dictionary('Titles')
        link_text = self.ref_parser.build_dictionary('Links')
        self.page_ids = []
        self.pages = {}
        for page_id, details in self.ref_parser.build_dictionaries('Page'):
            self.page_ids.append(page_id)
            self.pages[page_id] = details
        self.page_contents = {}
        for page_id, page in self.pages.items():
            self.page_contents[page_id] = self.ref_parser.get_section('PageContent:%s' % page_id)
            path = page.get('Content')
            if path:
                self.page_ids.remove(page_id)
            else:
                path = page['Path']
            self.paths[page_id] = path
            self.titles[page_id] = page.get('Title', page_id)
            link_text[page_id] = page.get('Link', self.titles[page_id])
        self.links = self._parse_links(link_text)

        self.other_code = self.ref_parser.build_dictionaries('OtherCode')
        self.code_path = self.get_code_path(code_id)
        self.game_vars = self.ref_parser.build_dictionary('Game')
        self.colours = self._get_colours(ref_parser.build_dictionary('Colours'))
        self.game = self.game_vars.get('Game', skoolfile[:-6])
        self.tap_info = self._get_tap_info()
        self.tap_address_table = self.tap_info.tap_address_table if self.tap_info else None
        self.map_details = self.ref_parser.build_dictionary('MapDetails')

        self.parser = SkoolParser(skoolfile, tap_info=self.tap_info)
        self.snapshot = self.parser.snapshot
        self._snapshots = [(self.snapshot, '')]
        self.entries = self.parser.entries
        self.memory_map = [e for e in self.parser.memory_map if e.ctl != 'i']
        self.tap_addresses = self.parser.tap_addresses
        self.tap_descs = self.parser.tap_descs

    def _get_colours(self, cdict):
        colours = {}
        for k, v in cdict.items():
            colours[k] = tuple([int(n) for n in v.split(',')])
        return colours

    def _get_tap_info(self):
        return None

    def _parse_links(self, links):
        new_links = {}
        for page_id, link_text in links.items():
            if link_text.startswith('['):
                sb_count = 1
                index = 1
                length = len(link_text)
                while index < length and sb_count > 0:
                    if link_text[index] == '[':
                        sb_count += 1
                    elif link_text[index] == ']':
                        sb_count -= 1
                    index += 1
                if sb_count == 0:
                    new_links[page_id] = (link_text[1:index - 1], link_text[index:])
            if page_id not in new_links:
                new_links[page_id] = (link_text.strip(), '')
        return new_links

    def get_code_path(self, code_id):
        for c_id, code in self.other_code:
            if c_id == code_id:
                return code['Path']
        return 'asm'

    def has_gbuffer(self):
        for entry in self.memory_map:
            if entry.ctl == 'g':
                return True
        return False

    def has_ref_items(self, section_type):
        return self.ref_parser.has_ref_items(section_type)

    def get_section(self, section_name):
        return self.ref_parser.get_section(section_name)

    def get_ref_items(self, section_type, keep_lines=False, as_list=False):
        return self.ref_parser.get_ref_items(section_type, keep_lines, as_list)

    def get_list(self, section_name, default=()):
        return self.ref_parser.get_list(section_name) or default

    def get_entry(self, address):
        return self.parser.get_entry(address)

    def get_entry_point_refs(self, address):
        return self.parser.get_entry_point_refs(address)

    def get_tap_refs(self, address):
        return self.parser.get_tap_refs(address)

    def get_container(self, address):
        return self.parser.get_container(address)

    def get_instruction_addr_str(self, address):
        return self.parser.instructions[address][0].addr_str

    def get_snapshot_name(self):
        return self._snapshots[-1][1]

    def pop_snapshot(self):
        self.snapshot = self._snapshots.pop()[0]

    def push_snapshot(self, name=''):
        self._snapshots.append((self.snapshot[:], name))

    def _get_screen_udg(self, row, col):
        attr = self.snapshot[22528 + 32 * row + col]
        address = 16384 + 2048 * (row // 8) + 32 * (row % 8) + col
        return Udg(attr, self.snapshot[address:address + 2048:256])

    def get_screenshot(self, x=0, y=0, w=32, h=24):
        width = min((w, 32 - x))
        height = min((h, 24 - y))
        scr_udgs = []
        for r in range(y, y + height):
            scr_udgs.append([self._get_screen_udg(r, c) for c in range(x, x + width)])
        return scr_udgs

    def _flip_byte(self, byte):
        f = 0
        for b in range(8):
            f = 2 * f + (byte & 1)
            byte /= 2
        return f

    def _flip_udg(self, udg):
        for index in range(8):
            udg.data[index] = self._flip_byte(udg.data[index])
            udg.mask[index] = self._flip_byte(udg.mask[index])

    def _get_sprite_udg(self, state, attr, ref_page, udg_page):
        ref_addr = state + 256 * ref_page
        ref = self.snapshot[ref_addr]
        if ref == 0:
            udg = [0] * 8
            mask = [255] * 8
        else:
            udg_addr = ref + 256 * udg_page
            udg = self.snapshot[udg_addr:udg_addr + 4096:512]
            mask = self.snapshot[udg_addr + 256:udg_addr + 4352:512]
        return Udg(attr, udg, mask, ref_addr=ref_addr, ref=ref, udg_page=udg_page)

    def _get_sprite_size(self, state):
        return 3, 4

    def build_sprite(self, state, attr=None, udg_page=None):
        width, height = self._get_sprite_size(state)
        udg_array = []
        for row in range(height):
            udg_array.append([self.get_sprite_udg(state, attr, row, col, udg_page) for col in range(width)])
        return udg_array

    def get_font_bitmap(self, character):
        base_address = ord(character) + self.font_address + 256
        width = self.snapshot[base_address - 256]
        return self.snapshot[base_address:base_address + 256 * width:256]

    def get_text(self, words):
        columns = []
        for character in words:
            columns.append(0)
            columns.extend(self.get_font_bitmap(character))
        udg_array = []
        for col in range(len(columns)):
            col_byte = columns[col]
            udg_index = col // 8
            bit = int(math.pow(2, 7 - (col - 8 * udg_index)))
            if udg_index == len(udg_array):
                udg_array.append([0] * 8)
            udg = udg_array[udg_index]
            for b in range(8):
                udg[7 - b] |= bit * (col_byte & 1)
                col_byte /= 2
        return udg_array

    def _draw_text(self, udgs, words, x, y):
        text_udgs = self.get_text(words)
        for col in range(len(text_udgs)):
            udg = udgs[y][col + x].data
            for index in range(8):
                udg[index] |= text_udgs[col][index]

    def get_skool_udgs(self, x, y, w, h, show_chars=False, show_x=0):
        skool_udgs = []
        for row in range(y, y + h):
            skool_udgs.append([])
            for col in range(x, x + w):
                skool_udgs[-1].append(self.get_skool_udg(row, col, show_chars))
        if show_x > 0:
            line = [Udg(8 * (7 - skool_x % 2), [0] * 8) for skool_x in range(x, x + w)]
            for skool_x in range(x, x + w):
                if skool_x % show_x == 0:
                    self._draw_text([line], str(skool_x), skool_x - x, 0)
            skool_udgs.insert(0, line)
            skool_udgs.append(line)
        return skool_udgs

    def _superimpose_sprite_udgs(self, skool_udg, x, y):
        for state, char_x, char_y in self.get_initial_states():
            if char_y <= y <= char_y + 3 and char_x <= x <= char_x + 2:
                sprite_udg = self.get_sprite_udg(state, None, y - char_y, x - char_x)
                for i in range(8):
                    skool_udg.data[i] = (skool_udg.data[i] | sprite_udg.data[i]) & sprite_udg.mask[i]

class SkoolDaze(SkoolGame):
    def __init__(self, skoolfile, ref_parser, code_id=MAIN_CODE_ID):
        SkoolGame.__init__(self, skoolfile, ref_parser, code_id)
        self.min_lesson = 224
        self.max_lesson = 255
        self.min_character = 152
        self.max_character = 169
        self.font_address = 55552
        self.char_buf_descs = self._get_char_buf_descs()
        self.keypress_routines = self._get_keypress_routines()
        self.as_descs = self._get_as_descs()
        self.characters = self._get_characters()
        self.lesson_descs = self._get_lesson_descs()

    def _get_char_buf_descs(self):
        char_buf_descs = []
        char_buf_descs.append(("96", (
            "Animatory state",
        )))
        char_buf_descs.append(("97,98", (
            "Location (y,x)",
        )))
        char_buf_descs.append(("99", (
            "LSB of the primary command routine address (see #R25126)",
            "ERIC's post-midstride animatory state (see #R25856, #R25916)",
        )))
        char_buf_descs.append(("100", (
            "MSB of the primary command routine address (see #R25126)",
            "ERIC's post-midstride y-coordinate (see #R25856, #R25916)",
        )))
        char_buf_descs.append(("101", (
            "Destination y-coordinate (see #R27090)",
            "Walkabout event identifier (see #R25303)",
            "ERIC's post-midstride x-coordinate (see #R25856, #R25916)",
        )))
        char_buf_descs.append(("102", (
            "Destination x-coordinate (see #R27090)",
            "ERIC's animatory state before he started firing the catapult, hitting, or jumping (see #R62574)",
        )))
        char_buf_descs.append(("103", (
            "Number of mini-walkabouts remaining (see #R27453)",
        )))
        char_buf_descs.append(("104", (
            "Message repetition delay counter (see #R63488)",
        )))
        char_buf_descs.append(("105,106", (
            "Interruptible subcommand routine address (see #R25126)",
        )))
        char_buf_descs.append(("107", (
            "LSB of the address of the next character in the message being written on a blackboard or spoken (see #R28882)",
            "Walkabout destination x-coordinate (see #R27442)",
            "x-coordinate increment to use while chasing ERIC (see #R31739, #R31768)",
            "Number of movements remaining to ascend or descend the staircase (see #R25488, #R25534)",
            "Number of movements remaining to wipe the blackboard (see #R29160)",
            "Intermediate destination x-coordinate (see #R25404)",
        )))
        char_buf_descs.append(("108", (
            "MSB of the address of the next character in the message being written on a blackboard or spoken (see #R28882)",
            "y-coordinate increment to use while chasing ERIC (see #R31739, #R31768)",
            "x-coordinate of the next column of the board to wipe (see #R29160)",
            "Number of paces remaining until the next command list restart request check (see #R25404)",
        )))
        char_buf_descs.append(("109", (
            "LSB of the address of the next character in the submessage being written on a blackboard or spoken (see #R28882)",
            "y-coordinate of the top row of the blackboard being wiped (see #R29160)",
        )))
        char_buf_descs.append(("110", (
            "MSB of the address of the next character in the submessage being written on a blackboard or spoken (see #R28882)",
        )))
        char_buf_descs.append(("111,112", (
            "Uninterruptible subcommand routine address (see #R25126)",
        )))
        char_buf_descs.append(("113", (
            "Counter determining how much longer the character will stay down after being punched (see #R28569), hit by a catapult pellet (see #R27932), knocked out of a chair (see #R27733), or tripped up by a stampeding kid (see #R27126)",
            "Distance remaining for the pellet to travel (see #R27932)",
            "ANGELFACE's pre-punch animatory state (see #R28446, #R28642)",
            "BOY WANDER's pre-fire animatory state (see #R28716, #R28786)",
        )))
        char_buf_descs.append(("114", (
            "Pre-knockout animatory state (see #R27206)",
        )))
        char_buf_descs.append(("115,116", (
            "Unused",
        )))
        char_buf_descs.append(("117,118", (
            "Address reached in the command list",
        )))
        char_buf_descs.append(("119,120", (
            "Start address of the command list",
        )))
        char_buf_descs.append(("121", (
            "Unused (always 0, but checked by #R25126)",
        )))
        char_buf_descs.append(("122", (
            """Various flags; initially populated by #R26880 with 0 for kids, 16 (bit 4 set) for teachers, and 32 (bit 5 set) for catapult pellets
#TABLE(bufDesc,bufBit)
{ =h Bit(s) | =h Meaning if set }
{ 0         | Restart the command list at the next opportunity if bit 1 is reset (see #R25126) }
{ 1         | Always reset }
{ 2-3       | Unused }
{ 4         | Character is a teacher (checked by #R25367) }
{ 5         | Character is walking fast continuously (set for pellets and stampeding boys; see #R27246) }
{ 6         | Character is walking slowly continuously (always reset, but checked by #R25367) }
{ 7         | Character is walking slowly (see #R25266) }
TABLE#""",
        )))
        char_buf_descs.append(("123", (
            "Walking speed change delay counter (see #R25266, #R32234, #R32242, #R64016)",
        )))
        char_buf_descs.append(("124,125", (
            "Continual subcommand routine address (see #R25126); set to #R25247 initially by #R26880",
        )))
        char_buf_descs.append(("126,127", (
            "Unused",
        )))
        return char_buf_descs

    def _get_keypress_routines(self):
        keypress_routines = {}
        keypress_routines[25983] = 'Down'
        keypress_routines[26084] = 'Up'
        keypress_routines[26194] = 'Left'
        keypress_routines[26282] = 'Right'
        keypress_routines[26449] = ''
        keypress_routines[27823] = 'Sit/stand'
        keypress_routines[62574] = 'Fire catapult'
        keypress_routines[62610] = 'Hit'
        keypress_routines[62617] = 'Jump'
        keypress_routines[63056] = 'Write'
        return keypress_routines

    def _get_as_descs(self):
        as_descs = {}
        as_descs[0] = 'ERIC standing/walking (phase 1)'
        as_descs[1] = 'ERIC walking (midstride) (phase 2)'
        as_descs[2] = 'ERIC standing/walking (phase 3)'
        as_descs[3] = 'ERIC walking (midstride) (phase 4)'
        as_descs[4] = 'ERIC sitting (as on a chair) with his hand raised (unused)'
        as_descs[5] = 'ERIC sitting (as on a chair or the stairs)'
        as_descs[6] = 'ERIC sitting (as on the floor)'
        as_descs[7] = 'ERIC lying on his back'
        as_descs[8] = 'ERIC firing his catapult (phase 1)'
        as_descs[9] = 'ERIC firing his catapult (phase 2)'
        as_descs[10] = 'ERIC hitting (phase 1)'
        as_descs[11] = 'ERIC hitting (phase 2)'
        as_descs[12] = 'ERIC with his arm up, as if jumping (phase 1)'
        as_descs[13] = 'ERIC with his arm up, as if writing or jumping (phase 2)'
        as_descs[14] = 'Unused'
        as_descs[15] = 'Unused'
        as_descs[16] = 'EINSTEIN standing/walking (phase 1)'
        as_descs[17] = 'EINSTEIN walking (midstride) (phase 2)'
        as_descs[18] = 'EINSTEIN standing/walking (phase 3)'
        as_descs[19] = 'EINSTEIN walking (midstride) (phase 4)'
        as_descs[20] = 'EINSTEIN sitting (as on a chair) with his hand raised (unused)'
        as_descs[21] = 'EINSTEIN sitting (as on a chair)'
        as_descs[22] = 'EINSTEIN sitting (as on the floor)'
        as_descs[23] = 'EINSTEIN lying on his back'
        as_descs[24] = 'Unused'
        as_descs[25] = 'Unused'
        as_descs[26] = 'Unused'
        as_descs[27] = 'Unused'
        as_descs[28] = 'Unused'
        as_descs[29] = 'Unused'
        as_descs[30] = 'Unused'
        as_descs[31] = 'Unused'
        as_descs[32] = 'ANGELFACE standing/walking (phase 1)'
        as_descs[33] = 'ANGELFACE walking (midstride) (phase 2)'
        as_descs[34] = 'ANGELFACE standing/walking (phase 3)'
        as_descs[35] = 'ANGELFACE walking (midstride) (phase 4)'
        as_descs[36] = 'ANGELFACE sitting (as on a chair) with his hand raised (unused)'
        as_descs[37] = 'ANGELFACE sitting (as on a chair)'
        as_descs[38] = 'ANGELFACE sitting (as on the floor)'
        as_descs[39] = 'ANGELFACE lying on his back'
        as_descs[40] = 'Unused'
        as_descs[41] = 'Unused'
        as_descs[42] = 'ANGELFACE hitting (phase 1)'
        as_descs[43] = 'ANGELFACE hitting (phase 2)'
        as_descs[44] = 'ANGELFACE with his arm up, as if writing (phase 1) (unused)'
        as_descs[45] = 'ANGELFACE with his arm up, as if writing (phase 2) (unused)'
        as_descs[46] = 'Unused'
        as_descs[47] = 'Unused'
        as_descs[48] = 'BOY WANDER standing/walking (phase 1)'
        as_descs[49] = 'BOY WANDER walking (midstride) (phase 2)'
        as_descs[50] = 'BOY WANDER standing/walking (phase 3)'
        as_descs[51] = 'BOY WANDER walking (midstride) (phase 4)'
        as_descs[52] = 'BOY WANDER sitting (as on a chair) with his hand raised (unused)'
        as_descs[53] = 'BOY WANDER sitting (as on a chair)'
        as_descs[54] = 'BOY WANDER sitting (as on the floor)'
        as_descs[55] = 'BOY WANDER lying on his back'
        as_descs[56] = 'BOY WANDER firing his catapult (phase 1)'
        as_descs[57] = 'BOY WANDER firing his catapult (phase 2)'
        as_descs[58] = 'Unused'
        as_descs[59] = 'Unused'
        as_descs[60] = 'BOY WANDER with his arm up, as if writing (phase 1)'
        as_descs[61] = 'BOY WANDER with his arm up, as if writing (phase 2)'
        as_descs[62] = 'Unused'
        as_descs[63] = 'BOY WANDER lying on his back (unused)'
        as_descs[64] = 'Boy standing/walking (phase 1)'
        as_descs[65] = 'Boy walking (midstride) (phase 2)'
        as_descs[66] = 'Boy standing/walking (phase 3)'
        as_descs[67] = 'Boy walking (midstride) (phase 4)'
        as_descs[68] = 'Boy sitting (as on a chair) (unused)'
        as_descs[69] = 'Boy sitting (as on a chair)'
        as_descs[70] = 'Boy sitting (as on the floor)'
        as_descs[71] = 'Boy lying on his back'
        as_descs[72] = 'MR CREAK standing/walking (phase 1)'
        as_descs[73] = 'MR CREAK walking (midstride) (phase 2)'
        as_descs[74] = 'MR CREAK standing/walking (phase 3)'
        as_descs[75] = 'MR CREAK walking (midstride) (phase 4)'
        as_descs[76] = 'MR CREAK with his arm up (unused)'
        as_descs[77] = 'MR CREAK with his arm up'
        as_descs[78] = 'MR CREAK sitting (as on the floor) (unused)'
        as_descs[79] = 'MR CREAK sitting (as on the floor)'
        as_descs[80] = 'MR WACKER standing/walking (phase 1)'
        as_descs[81] = 'MR WACKER walking (midstride) (phase 2)'
        as_descs[82] = 'MR WACKER standing/walking (phase 3)'
        as_descs[83] = 'MR WACKER walking (midstride) (phase 4)'
        as_descs[84] = 'MR WACKER with his arm up (unused)'
        as_descs[85] = 'MR WACKER with his arm up'
        as_descs[86] = 'MR WACKER sitting (as on the floor) (unused)'
        as_descs[87] = 'MR WACKER sitting (as on the floor)'
        as_descs[88] = 'MR ROCKITT standing/walking (phase 1)'
        as_descs[89] = 'MR ROCKITT walking (midstride) (phase 2)'
        as_descs[90] = 'MR ROCKITT standing/walking (phase 3)'
        as_descs[91] = 'MR ROCKITT walking (midstride) (phase 4)'
        as_descs[92] = 'MR ROCKITT with his arm up (unused)'
        as_descs[93] = 'MR ROCKITT with his arm up'
        as_descs[94] = 'MR ROCKITT sitting (as on the floor) (unused)'
        as_descs[95] = 'MR ROCKITT sitting (as on the floor)'
        as_descs[96] = 'MR WITHIT standing/walking (phase 1)'
        as_descs[97] = 'MR WITHIT walking (midstride) (phase 2)'
        as_descs[98] = 'MR WITHIT standing/walking (phase 3)'
        as_descs[99] = 'MR WITHIT walking (midstride) (phase 4)'
        as_descs[100] = 'MR WITHIT with his arm up (unused)'
        as_descs[101] = 'MR WITHIT with his arm up'
        as_descs[102] = 'MR WITHIT sitting (as on the floor) (unused)'
        as_descs[103] = 'MR WITHIT sitting (as on the floor)'
        as_descs[104] = 'Unused'
        as_descs[105] = 'Unused'
        as_descs[106] = 'Unused'
        as_descs[107] = 'Unused'
        as_descs[108] = 'Unused'
        as_descs[109] = 'Unused'
        as_descs[110] = 'Unused'
        as_descs[111] = 'Unused'
        as_descs[112] = 'Unused'
        as_descs[113] = 'Unused'
        as_descs[114] = 'Unused'
        as_descs[115] = 'Unused'
        as_descs[116] = 'Unused'
        as_descs[117] = 'Unused'
        as_descs[118] = 'Unused'
        as_descs[119] = 'Unused'
        as_descs[120] = 'Unused'
        as_descs[121] = 'Unused'
        as_descs[122] = 'Unused'
        as_descs[123] = 'Unused'
        as_descs[124] = 'Unused'
        as_descs[125] = 'Unused'
        as_descs[126] = 'Unused'
        as_descs[127] = 'Catapult pellet'
        return as_descs

    def _get_characters(self):
        characters = {}
        characters[152] = "Little boy no. 1"
        characters[153] = "Little boy no. 2"
        characters[154] = "Little boy no. 3"
        characters[155] = "Little boy no. 4"
        characters[156] = "Little boy no. 5"
        characters[157] = "Little boy no. 6"
        characters[158] = "Little boy no. 7"
        characters[159] = "Little boy no. 8"
        characters[160] = "Little boy no. 9"
        characters[161] = "Little boy no. 10"
        characters[162] = "Little boy no. 11"
        characters[163] = "MR WACKER"
        characters[164] = "MR ROCKITT"
        characters[165] = "MR WITHIT"
        characters[166] = "MR CREAK"
        characters[167] = "BOY WANDER"
        characters[168] = "ANGELFACE"
        characters[169] = "EINSTEIN"
        characters[170] = "BOY WANDER's pellet"
        characters[171] = "ERIC's pellet"
        characters[172] = "ERIC"
        return characters

    def _get_lesson_descs(self):
        lesson_descs = {}
        lesson_descs[224] = 'DINNER (MR WITHIT)'
        lesson_descs[225] = 'DINNER (MR WACKER)'
        lesson_descs[226] = 'MR WACKER - EXAM ROOM'
        lesson_descs[227] = 'MR ROCKITT - EXAM ROOM'
        lesson_descs[228] = 'REVISION LIBRARY'
        lesson_descs[229] = 'REVISION LIBRARY'
        lesson_descs[230] = 'REVISION LIBRARY'
        lesson_descs[231] = 'MR WITHIT - MAP ROOM'
        lesson_descs[232] = 'MR WACKER - READING ROOM'
        lesson_descs[233] = 'MR ROCKITT - READING ROOM'
        lesson_descs[234] = 'MR CREAK - READING ROOM'
        lesson_descs[235] = 'MR CREAK - WHITE ROOM'
        lesson_descs[236] = 'MR WACKER - WHITE ROOM'
        lesson_descs[237] = 'MR WITHIT - WHITE ROOM'
        lesson_descs[238] = 'MR ROCKITT - WHITE ROOM'
        lesson_descs[239] = 'MR WACKER - MAP ROOM'
        lesson_descs[240] = 'MR WITHIT - MAP ROOM'
        lesson_descs[241] = 'MR ROCKITT - WHITE ROOM'
        lesson_descs[242] = 'MR CREAK - READING ROOM'
        lesson_descs[243] = 'PLAYTIME (ANGELFACE has mumps)'
        lesson_descs[244] = 'PLAYTIME (EINSTEIN is going to grass on ERIC)'
        lesson_descs[245] = 'PLAYTIME (BOY WANDER has hidden a pea-shooter)'
        lesson_descs[246] = 'PLAYTIME'
        lesson_descs[247] = 'PLAYTIME'
        lesson_descs[248] = 'PLAYTIME'
        lesson_descs[249] = 'PLAYTIME'
        lesson_descs[250] = 'PLAYTIME'
        lesson_descs[251] = 'PLAYTIME'
        lesson_descs[252] = 'PLAYTIME (little boys stampede)'
        lesson_descs[253] = 'PLAYTIME (little boys stampede)'
        lesson_descs[254] = 'PLAYTIME'
        lesson_descs[255] = 'PLAYTIME'
        return lesson_descs

    def _get_tap_info(self):
        return TapInfo(65152, 222)

    def get_skool_udg(self, y, x, show_chars=False):
        ref_page = y + 152
        attr_addr = x + 128 + 256 * ref_page
        ref_addr = x + 256 * ref_page
        ref = self.snapshot[ref_addr]
        attr = self.snapshot[attr_addr]
        udg_page = 128 + 8 * (x // 32)
        udg_addr = ref + 256 * udg_page
        udg_bytes = self.snapshot[udg_addr:udg_addr + 2048:256]
        skool_udg = Udg(attr, udg_bytes, attr_addr=attr_addr, ref_addr=ref_addr, ref=ref, udg_page=udg_page, x=x, y=y)
        if show_chars:
            self._superimpose_sprite_udgs(skool_udg, x, y)
        return skool_udg

    def get_sprite_udg(self, state, attr, row, col, udg_page=None):
        if udg_page is None:
            udg_page = 201 if 71 < (state & 127) < 104 else 185
        ref_page = 173 + col * 4 + row
        return self._get_sprite_udg(state, attr, ref_page, udg_page)

    def get_initial_states(self):
        initial_states = []
        for c in range(18) + [20]:
            initial_states.append((self.snapshot[56064 + c], self.snapshot[56320 + c], self.snapshot[56576 + c]))
        return initial_states

class BackToSkool(SkoolGame):
    def __init__(self, skoolfile, ref_parser, code_id=MAIN_CODE_ID):
        SkoolGame.__init__(self, skoolfile, ref_parser, code_id)
        self.min_lesson = 37
        self.max_lesson = 59
        self.min_character = 183
        self.max_character = 209
        self.font_address = 55040
        self.char_buf_descs = self._get_char_buf_descs()
        self.keypress_routines = self._get_keypress_routines()
        self.as_descs = self._get_as_descs()
        self.characters = self._get_characters()
        self.lesson_descs = self._get_lesson_descs()

    def _get_char_buf_descs(self):
        char_buf_descs = []
        char_buf_descs.append(("0", (
            "Animatory state",
        )))
        char_buf_descs.append(("1,2", (
            "Location (x, y)",
        )))
        char_buf_descs.append(("3", (
            "LSB of the primary command routine address (see #R25296)",
            "ERIC's post-midstride animatory state (see #R28114, #R28160)",
        )))
        char_buf_descs.append(("4", (
            "MSB of the primary command routine address (see #R25296)",
            "ERIC's post-midstride x-coordinate (see #R28114, #R28160)",
        )))
        char_buf_descs.append(("5", (
            "Destination x-coordinate (see #R25700)",
            "ERIC's post-midstride y-coordinate (see #R28114, #R28160)",
        )))
        char_buf_descs.append(("6", (
            "Destination y-coordinate (see #R25700)",
            "Walkabout event identifier (see #R31969)",
        )))
        char_buf_descs.append(("7", (
            "Number of mini-walkabouts remaining (see #R31969)",
        )))
        char_buf_descs.append(("8", (
            "Unused",
        )))
        char_buf_descs.append(("9,10", (
            "Interruptible subcommand routine address (see #R25296)",
        )))
        char_buf_descs.append(("11", (
            "LSB of the address of the next character in the message being written into a graphic buffer (buffer #R54784(214) only; see #R27419)",
            "LSB of the address of the next character in the message being written on a blackboard or spoken (see #R26318)",
            "Number of steps left to be ascended (see #R25656) or descended (see #R25677)",
            "Number of movements remaining to wipe the blackboard (see #R26639)",
            "Intermediate destination x-coordinate (see #R25581)",
        )))
        char_buf_descs.append(("12", (
            "MSB of the address of the next character in the message being written into a graphic buffer (buffer #R54784(214) only; see #R27419)",
            "MSB of the address of the next character in the message being written on a blackboard or spoken (see #R26318)",
            "Number of paces remaining until the next command list restart request check (see #R25581)",
            "x-coordinate of the column of the blackboard being wiped (see #R26639)",
        )))
        char_buf_descs.append(("13", (
            "LSB of the address of the next character in the submessage being written on a blackboard or spoken (see #R26318)",
            "y-coordinate of the top row of the blackboard being wiped (see #R26639)",
        )))
        char_buf_descs.append(("14", (
            "MSB of the address of the next character in the submessage being written on a blackboard or spoken (see #R26318)",
            "Identifier of the blackboard being wiped (see #R26639)",
        )))
        char_buf_descs.append(("15", (
            "LSB of the address of the next character in the sub-submessage being written on a blackboard or spoken (see #R26318)",
            "Animatory state of the character whose buffer is being borrowed by a mouse (see #R31254, #R31502)",
            "Post-midstride y-coordinate increment to use while chasing ERIC (see #R25944)",
        )))
        char_buf_descs.append(("16", (
            "MSB of the address of the next character in the sub-submessage being written on a blackboard or spoken (see #R26318)",
            "Counter determining the lifespan of a mouse released by ERIC (see #R31254, #R31502)",
        )))
        char_buf_descs.append(("17,18", (
            "Uninterruptible subcommand routine address (see #R25296)",
        )))
        char_buf_descs.append(("19", (
            "Identifier of the door or window being opened or closed (see #R29052)",
            "MR WACKER's x-coordinate before he goes to open a window (see #R63586)",
            "HAYLEY's pre-kiss x-coordinate (see #R24328, #R24348)",
            "Distance remaining for the mouse to travel before hiding or considering a change of direction (see #R31254)",
            "Counter determining how much longer the character will stay down after being knocked over (see #R30102) or knocked out of a chair (see #R26224)",
            "Distance remaining for the pellet to travel (see #R30380)",
            "Counter determining how much longer the female character should continue jumping up and down (see #R31078)",
            "Counter determining how much longer the desk lid should stay up (see #R62178)",
            "Counter determining how much longer the stinkbomb cloud will stick around (see #R63586)",
            "Water/sherry animation phase identifier (see #R63915)",
            "Counter determining the phases of growth of a watered plant (see #R64077)",
        )))
        char_buf_descs.append(("20", (
            "Open/close parameter for the door or window being opened or closed (see #R29052)",
            "MR WACKER's animatory state before he goes to open a window (see #R63586)",
            "HAYLEY's pre-kiss animatory state (see #R24328, #R24348)",
            "Counter determining when the mouse should hide (see #R31254)",
            "Pre-knockout animatory state (see #R30102)",
            "ANGELFACE's pre-punch animatory state (see #R30739)",
            "BOY WANDER's pre-fire animatory state (see #R30555)",
            "x-coordinate at which the plant will appear when it starts growing (see #R64077)",
        )))
        char_buf_descs.append(("21", (
            "Counter determining when the mouse should come out of hiding (see #R31254, #R31502)",
            "0 if the frightened female is jumping up and down, or 1 if she's standing on a chair (see #R31092)",
            "x-coordinate of the window MR WACKER will open after smelling a stinkbomb (see #R63586, #R63801)",
        )))
        char_buf_descs.append(("22", (
            "x-coordinate of the mouse when it last hid (see #R31254, #R31502)",
            "Identifier of the window MR WACKER will open after smelling a stinkbomb (see #R63586, #R63801)",
        )))
        char_buf_descs.append(("23,24", (
            "Continual subcommand routine address (see #R25296)",
        )))
        char_buf_descs.append(("25,26", (
            "Address reached in the command list",
        )))
        char_buf_descs.append(("27,28", (
            "Start address of the command list",
        )))
        char_buf_descs.append(("29", (
            """Various flags; initially populated by #R62910 with a copy of byte 35 of the character's buffer
#TABLE(bufDesc,bufBit)
{ =h Bit(s) | =h Meaning if set }
{ 0         | Restart the command list at the next opportunity (see #R25296) }
{ 1-2       | Unused }
{ 3         | Character is a candidate for immediate transportation to his destination (set by #R63309; checked and reset by #R25134; also reset by #R25296) }
{ 4         | Unused (always reset, but checked by #R25296; if set, it would make the character run until the current primary command has terminated) }
{ 5         | Character is an adult (see #R25296) }
{ 6         | Character is not human (see #R25296) }
{ 7         | Character is running (see #R25296) }
TABLE#""",
        )))
        char_buf_descs.append(("30", (
            "Walking speed change delay counter (see #R25296); initialised to bits 0-3 of the character number (see #R62910)",
        )))
        char_buf_descs.append(("31", (
            "Unused",
        )))
        char_buf_descs.append(("32", (
            "Initial animatory state (see #R62910)",
        )))
        char_buf_descs.append(("33,34", (
            "Initial location (see #R62910)",
        )))
        char_buf_descs.append(("35", (
            "Initial flags for byte 29 (see #R62910)",
        )))
        char_buf_descs.append(("36", (
            "Random location table identifier (see #R25815)",
        )))
        return char_buf_descs

    def _get_keypress_routines(self):
        keypress_routines = {}
        keypress_routines[23863] = 'Throw away water pistol'
        keypress_routines[23887] = 'Jump'
        keypress_routines[24133] = 'Write'
        keypress_routines[24263] = 'Hit'
        keypress_routines[24296] = 'Fire catapult'
        keypress_routines[24348] = 'Kiss'
        keypress_routines[28436] = 'Right'
        keypress_routines[28460] = 'Left'
        keypress_routines[28484] = 'Up'
        keypress_routines[28576] = 'Down'
        keypress_routines[28992] = 'Drop stinkbomb'
        keypress_routines[29574] = 'Mount bike'
        keypress_routines[57637] = 'Catch mouse/frog'
        keypress_routines[57893] = 'Release mice'
        keypress_routines[62423] = 'Sit'
        keypress_routines[63898] = 'Fire water pistol'
        return keypress_routines

    def _get_as_descs(self):
        as_descs = {}
        as_descs[0] = 'ERIC standing/walking (phase 1)'
        as_descs[1] = 'ERIC walking (midstride) (phase 2)'
        as_descs[2] = 'ERIC standing/walking (phase 3)'
        as_descs[3] = 'ERIC walking (midstride) (phase 4)'
        as_descs[4] = 'ERIC sitting (as on a chair)'
        as_descs[5] = 'ERIC sitting (as on the floor)'
        as_descs[6] = 'ERIC lying on his back'
        as_descs[7] = 'ERIC with his arm up (as if writing or jumping)'
        as_descs[8] = 'ERIC hitting (phase 1)'
        as_descs[9] = 'ERIC hitting (phase 2)'
        as_descs[10] = 'ERIC firing his catapult (phase 1)'
        as_descs[11] = 'ERIC firing his catapult (phase 2)'
        as_descs[12] = 'ERIC riding the bike (phase 1)'
        as_descs[13] = 'ERIC riding the bike (phase 2)'
        as_descs[14] = 'ERIC firing the water pistol'
        as_descs[15] = 'ERIC kissing HAYLEY'
        as_descs[16] = 'BOY WANDER standing/walking (phase 1)'
        as_descs[17] = 'BOY WANDER walking (midstride) (phase 2)'
        as_descs[18] = 'BOY WANDER standing/walking (phase 3)'
        as_descs[19] = 'BOY WANDER walking (midstride) (phase 4)'
        as_descs[20] = 'BOY WANDER sitting (as on a chair)'
        as_descs[21] = 'BOY WANDER sitting (as on the floor)'
        as_descs[22] = 'BOY WANDER lying on his back'
        as_descs[23] = 'BOY WANDER with his arm up (as if writing)'
        as_descs[24] = 'Bike lying on the floor'
        as_descs[25] = 'Bike standing upright'
        as_descs[26] = 'BOY WANDER firing his catapult (phase 1)'
        as_descs[27] = 'BOY WANDER firing his catapult (phase 2)'
        as_descs[28] = 'Frog sitting on the floor'
        as_descs[29] = 'Frog in the air (phase 1)'
        as_descs[30] = 'Frog in the air (phase 2)'
        as_descs[31] = 'Drop of water/sherry falling from a cup'
        as_descs[32] = 'ANGELFACE standing/walking (phase 1)'
        as_descs[33] = 'ANGELFACE walking (midstride) (phase 2)'
        as_descs[34] = 'ANGELFACE standing/walking (phase 3)'
        as_descs[35] = 'ANGELFACE walking (midstride) (phase 4)'
        as_descs[36] = 'ANGELFACE sitting (as on a chair)'
        as_descs[37] = 'ANGELFACE sitting (as on the floor)'
        as_descs[38] = 'ANGELFACE lying on his back'
        as_descs[39] = 'ERIC bending over'
        as_descs[40] = 'ANGELFACE hitting (phase 1)'
        as_descs[41] = 'ANGELFACE hitting (phase 2)'
        as_descs[42] = 'Flower half-grown'
        as_descs[43] = 'Flower fully grown'
        as_descs[44] = 'Desk lid (empty)'
        as_descs[45] = 'Desk lid (with pistol)'
        as_descs[46] = 'Desk lid (with stinkbombs)'
        as_descs[47] = 'Free mouse'
        as_descs[48] = 'EINSTEIN standing/walking (phase 1)'
        as_descs[49] = 'EINSTEIN walking (midstride) (phase 2)'
        as_descs[50] = 'EINSTEIN standing/walking (phase 3)'
        as_descs[51] = 'EINSTEIN walking (midstride) (phase 4)'
        as_descs[52] = 'EINSTEIN sitting (as on a chair)'
        as_descs[53] = 'EINSTEIN sitting (as on the floor)'
        as_descs[54] = 'EINSTEIN lying on his back'
        as_descs[55] = 'Conker'
        as_descs[56] = 'HAYLEY standing/walking (phase 1)'
        as_descs[57] = 'HAYLEY walking (midstride) (phase 2)'
        as_descs[58] = 'HAYLEY standing/walking (phase 3)'
        as_descs[59] = 'HAYLEY walking (midstride) (phase 4)'
        as_descs[60] = 'HAYLEY sitting (as on a chair)'
        as_descs[61] = 'HAYLEY sitting (as on the floor)'
        as_descs[62] = 'HAYLEY lying on her back'
        as_descs[63] = 'HAYLEY with her arm up (as if hitting Eric)'
        as_descs[64] = 'Boy standing/walking (phase 1)'
        as_descs[65] = 'Boy walking (midstride) (phase 2)'
        as_descs[66] = 'Boy standing/walking (phase 3)'
        as_descs[67] = 'Boy walking (midstride) (phase 4)'
        as_descs[68] = 'Boy sitting (as on a chair)'
        as_descs[69] = 'Boy sitting (as on the floor)'
        as_descs[70] = 'Boy lying on his back'
        as_descs[71] = 'Stinkbomb cloud phases 1 and 2'
        as_descs[72] = 'Girl standing/walking (phase 1)'
        as_descs[73] = 'Girl walking (midstride) (phase 2)'
        as_descs[74] = 'Girl standing/walking (phase 3)'
        as_descs[75] = 'Girl walking (midstride) (phase 4)'
        as_descs[76] = 'Girl sitting (as on a chair)'
        as_descs[77] = 'Girl sitting (as on the floor)'
        as_descs[78] = 'Girl lying on her back'
        as_descs[79] = 'Catapult pellet'
        as_descs[80] = 'MR WACKER standing/walking (phase 1)'
        as_descs[81] = 'MR WACKER walking (midstride) (phase 2)'
        as_descs[82] = 'MR WACKER standing/walking (phase 3)'
        as_descs[83] = 'MR WACKER walking (midstride) (phase 4)'
        as_descs[84] = 'Water fired from the pistol (phase 1)'
        as_descs[85] = 'Unused'
        as_descs[86] = 'MR WACKER sitting (as on the floor)'
        as_descs[87] = 'MR WACKER with his arm up'
        as_descs[88] = 'MR WITHIT standing/walking (phase 1)'
        as_descs[89] = 'MR WITHIT walking (midstride) (phase 2)'
        as_descs[90] = 'MR WITHIT standing/walking (phase 3)'
        as_descs[91] = 'MR WITHIT walking (midstride) (phase 4)'
        as_descs[92] = 'Water fired from the pistol (phase 2)'
        as_descs[93] = 'Unused'
        as_descs[94] = 'MR WITHIT sitting (as on the floor)'
        as_descs[95] = 'MR WITHIT with his arm up'
        as_descs[96] = 'MR ROCKITT standing/walking (phase 1)'
        as_descs[97] = 'MR ROCKITT walking (midstride) (phase 2)'
        as_descs[98] = 'MR ROCKITT standing/walking (phase 3)'
        as_descs[99] = 'MR ROCKITT walking (midstride) (phase 4)'
        as_descs[100] = 'Water fired from the pistol (unused)'
        as_descs[101] = 'Unused'
        as_descs[102] = 'MR ROCKITT sitting (as on the floor)'
        as_descs[103] = 'MR ROCKITT with his arm up'
        as_descs[104] = 'MR CREAK standing/walking (phase 1)'
        as_descs[105] = 'MR CREAK walking (midstride) (phase 2)'
        as_descs[106] = 'MR CREAK standing/walking (phase 3)'
        as_descs[107] = 'MR CREAK walking (midstride) (phase 4)'
        as_descs[108] = 'Water fired from the pistol (phase 3)'
        as_descs[109] = 'Unused'
        as_descs[110] = 'MR CREAK sitting (as on the floor)'
        as_descs[111] = 'MR CREAK with his arm up'
        as_descs[112] = 'MISS TAKE standing/walking (phase 1)'
        as_descs[113] = 'MISS TAKE walking (midstride) (phase 2)'
        as_descs[114] = 'MISS TAKE standing/walking (phase 3)'
        as_descs[115] = 'MISS TAKE walking (midstride) (phase 4)'
        as_descs[116] = 'Water fired from the pistol (phase 4)'
        as_descs[117] = 'Unused'
        as_descs[118] = 'MISS TAKE sitting (as on the floor)'
        as_descs[119] = 'MISS TAKE with her arm up'
        as_descs[120] = 'ALBERT standing/walking (phase 1)'
        as_descs[121] = 'ALBERT walking (midstride) (phase 2)'
        as_descs[122] = 'ALBERT standing/walking (phase 3)'
        as_descs[123] = 'ALBERT walking (midstride) (phase 4)'
        as_descs[124] = 'Water fired from the pistol (phase 5+)'
        as_descs[125] = 'Unused'
        as_descs[126] = 'ALBERT lying on his back'
        as_descs[127] = 'ALBERT with his arm up'
        return as_descs

    def _get_characters(self):
        characters = {}
        characters[183] = "Little girl no. 1"
        characters[184] = "Little girl no. 2"
        characters[185] = "Little girl no. 3"
        characters[186] = "Little girl no. 4"
        characters[187] = "Little girl no. 5"
        characters[188] = "Little girl no. 6"
        characters[189] = "Little girl no. 7"
        characters[190] = "Little boy no. 1"
        characters[191] = "Little boy no. 2"
        characters[192] = "Little boy no. 3"
        characters[193] = "Little boy no. 4"
        characters[194] = "Little boy no. 5"
        characters[195] = "Little boy no. 6"
        characters[196] = "Little boy no. 7"
        characters[197] = "Little boy no. 8"
        characters[198] = "Little boy no. 9"
        characters[199] = "Little boy no. 10"
        characters[200] = "MR WACKER"
        characters[201] = "MR WITHIT"
        characters[202] = "MR ROCKITT"
        characters[203] = "MR CREAK"
        characters[204] = "MISS TAKE"
        characters[205] = "ALBERT"
        characters[206] = "BOY WANDER"
        characters[207] = "ANGELFACE"
        characters[208] = "EINSTEIN"
        characters[209] = "HAYLEY"
        characters[210] = "ERIC"
        characters[211] = "Bike"
        characters[212] = "Frog; mouse"
        characters[213] = "BOY WANDER's pellet; stinkbomb cloud"
        characters[214] = "ERIC's pellet; water/sherry; plant; conker; desk lid"
        return characters

    def _get_lesson_descs(self):
        lesson_descs = {}
        lesson_descs[37] = 'MR CREAK - BLUE ROOM'
        lesson_descs[38] = 'MR CREAK - BLUE ROOM'
        lesson_descs[39] = 'MR CREAK - YELLOW ROOM'
        lesson_descs[40] = 'MR CREAK - YELLOW ROOM'
        lesson_descs[41] = 'MR WITHIT - BLUE ROOM'
        lesson_descs[42] = 'MR WITHIT - BLUE ROOM'
        lesson_descs[43] = 'MR WITHIT - YELLOW ROOM'
        lesson_descs[44] = 'MR WITHIT - YELLOW ROOM'
        lesson_descs[45] = 'MR WITHIT - SCIENCE LAB'
        lesson_descs[46] = 'MR WITHIT - SCIENCE LAB'
        lesson_descs[47] = 'MR ROCKITT - SCIENCE LAB'
        lesson_descs[48] = 'MR ROCKITT - SCIENCE LAB'
        lesson_descs[49] = 'MR ROCKITT - SCIENCE LAB'
        lesson_descs[50] = 'DINNER (MR WACKER)'
        lesson_descs[51] = 'DINNER (MR WITHIT)'
        lesson_descs[52] = 'ASSEMBLY'
        lesson_descs[53] = 'REVISION LIBRARY'
        lesson_descs[54] = 'REVISION LIBRARY'
        lesson_descs[55] = 'PLAYTIME'
        lesson_descs[56] = 'PLAYTIME'
        lesson_descs[57] = 'PLAYTIME'
        lesson_descs[58] = 'PLAYTIME'
        lesson_descs[59] = 'PLAYTIME'
        return lesson_descs

    def _get_tap_info(self):
        return TapInfo(59392, 88)

    def get_skool_udg(self, y, x, show_chars=False):
        q_addr = x + 46336
        q = self.snapshot[q_addr]
        h = y + 160
        hl = 180 + q // 2 + 256 * h
        i = 0
        if q & 1:
            p = self.snapshot[hl] & 15
        else:
            p = (self.snapshot[hl] & 240) // 16
        if p == 0:
            i = q & 3 or 2
            if q & 1:
                p = self.snapshot[hl + 256] & 15
            else:
                p = (self.snapshot[hl + 256] & 240) // 16
        attr = i | (8 * p)
        ref_addr = q + 256 * h
        ref = self.snapshot[ref_addr]
        a = 136 // int(math.pow(2, q & 3)) & self.snapshot[144 + q // 4 + 256 * h]
        udg_page = (8 if a & 15 else 0) + (144 if a & 240 else 128)
        udg_addr = ref + 256 * udg_page
        udg_bytes = self.snapshot[udg_addr:udg_addr + 2048:256]
        skool_udg = Udg(attr, udg_bytes, q_addr=q_addr, ref_addr=ref_addr, ref=ref, udg_page=udg_page, x=x, y=y)
        if show_chars:
            self._superimpose_sprite_udgs(skool_udg, x, y)
        return skool_udg

    def get_sprite_udg(self, state, attr, row, col, udg_page=None):
        as_norm = state | 128
        udg_page = 183 if as_norm < 208 else 199
        col = 2 - col if state & 128 else col
        ref_page = 215 + col * 4 + row
        udg = self._get_sprite_udg(as_norm, attr, ref_page, udg_page)
        if state & 128:
            self._flip_udg(udg)
        return udg

    def _alter_skool_udgs(self, address):
        coords = []
        while True:
            y = self.snapshot[address]
            if y == 255:
                break
            x = self.snapshot[address + 1]
            coords.append((x, y))
            udg_lsb = self.snapshot[address + 2]
            attr = self.snapshot[address + 3]
            d = y + 160
            col_ref = self.snapshot[x + 46336]
            self.snapshot[col_ref + 256 * d] = udg_lsb
            z = int(math.pow(2, col_ref & 3))
            e = 128 + (col_ref + 104) // 2
            a = self.snapshot[e + 256 * d]
            if col_ref & 1:
                a = (a & 240) | (attr & 15)
            else:
                a = (16 * (attr & 15)) | (a & 15)
            self.snapshot[e + 256 * d] = a
            a = attr & 192
            if a & 64:
                a = a - 56
            e = 144 + col_ref // 4
            self.snapshot[e + 256 * d] = self.snapshot[e + 256 * d] & 255 - 136 // z | a // z
            address += 4
        return coords

    def _get_skool_udgs_by_coords(self, coords):
        left_x, top_y, width, height = self._get_dimensions(coords)
        udg_array = []
        for y in range(top_y, top_y + height):
            udg_array.append([None] * width)
        for x, y in coords:
            udg_array[y - top_y][x - left_x] = self.get_skool_udg(y, x)
        return udg_array

    def _get_dimensions(self, coords, padding=0):
        x_range = [c[0] for c in coords]
        y_range = [c[1] for c in coords]
        left_x = min(x_range)
        top_y = min(y_range)
        width = max(x_range) - left_x + 1
        height = max(y_range) - top_y + 1
        left_x -= padding
        top_y -= padding
        width += 2 * padding if left_x + width < 192 else padding
        height += 2 * padding if top_y + height < 20 else padding
        return left_x, top_y, width, height

    def get_udg(self, address, attr):
        return Udg(attr, self.snapshot[address:address + 8])

    def get_mutable_udg_array(self, address, padding):
        self.push_snapshot()
        coords = self._alter_skool_udgs(address)
        if padding >= 0:
            udg_array = self.get_skool_udgs(*self._get_dimensions(coords, padding))
        else:
            udg_array = self._get_skool_udgs_by_coords(coords)
        self.pop_snapshot()
        return udg_array

    def get_initial_states(self):
        return [self.snapshot[32 + 256 * c:35 + 256 * c] for c in range(183, 211)]

class ContactSamCruise(SkoolGame):
    def __init__(self, skoolfile, ref_parser, code_id=MAIN_CODE_ID):
        SkoolGame.__init__(self, skoolfile, ref_parser, code_id)
        self.font = {
            48: (124, 138, 146, 124),
            49: (66, 254, 2),
            50: (70, 138, 146, 98),
            51: (130, 146, 178, 204),
            52: (56, 72, 254, 8),
            53: (242, 146, 146, 140),
            54: (124, 146, 146, 140),
            55: (128, 134, 152, 224),
            56: (108, 146, 146, 108),
            57: (98, 146, 146, 124)
        }
        self.char_buf_descs = self._get_char_buf_descs()
        self.keypress_routines = self._get_keypress_routines()
        self.as_descs = self._get_as_descs()
        self.characters = self._get_characters()

    def _get_char_buf_descs(self):
        char_buf_descs = []
        char_buf_descs.append(("0", (
            "Animatory state",
        )))
        char_buf_descs.append(("1,2", (
            "Location (x, y)",
        )))
        char_buf_descs.append(("3", (
            """Various flags
#TABLE(bufDesc,bufBit)
{ =h Bit(s) | =h Meaning if set }
{ 0-3       | Unknown }
{ 4         | Restart the command list at the next opportunity (see #R61948) }
{ 5         | Unknown }
{ 6         | Character should be moved (see #R61948) }
{ 7         | Character is running (see #R61948) }
TABLE#""",
        )))
        char_buf_descs.append(("4", (
            "z-coordinate: 1=indoors; 2=outside, behind people on sidewalk; 4=on sidewalk or road; 8=invisible",
        )))
        char_buf_descs.append(("5", (
            "Post-midstride animatory state (see #R60347, #R60506)",
        )))
        char_buf_descs.append(("6", (
            "Post-midstride x-coordinate  (see #R60347, #R60506)",
        )))
        char_buf_descs.append(("7", (
            "Post-midstride y-coordinate  (see #R60347, #R60506)",
            "Unknown (see #R60641, #R64338)",
        )))
        char_buf_descs.append(("8", (
            "LSB of the primary command routine address (see #R61948)",
            "Unknown (see #R24832, #R30675, #R60641, #R64338)",
        )))
        char_buf_descs.append(("9", (
            "MSB of the primary command routine address (see #R61948)",
            "Unknown (see #R30989, #R60641)",
        )))
        char_buf_descs.append(("10", (
            "Destination x-coordinate (see #R31944)",
            "Walkabout origin x-coordinate (see #R63304, #R63530)",
            "Unknown (see #R25420)",
        )))
        char_buf_descs.append(("11", (
            "Destination y-coordinate (see #R31944)",
            "Walkabout destination x-coordinate (see #R63530)",
            "Unknown (see #R31338, #R63304, #R63520)",
        )))
        char_buf_descs.append(("12", (
            "Roll/somersault animation phase table entry address LSB (see #R25789)",
            "Destination location identifier (see #R29072)",
            "Unknown (see #R28803, #R31338, #R31808, #R60812, #R63304, #R63520)",
        )))
        char_buf_descs.append(("13", (
            "Roll/somersault animation phase table entry address MSB (see #R25789)",
            "0 if Sam has only just met Lana, 1 otherwise (see #R24832)",
            "Pre-knockout animatory state (see #R64005)",
            "Unknown (see #R31338, #R31808, #R60812, #R62976, #R63304, #R64227)",
        )))
        char_buf_descs.append(("14,15", (
            "Interruptible subcommand routine address (see #R61948)",
        )))
        char_buf_descs.append(("16", (
            "Walkabout destination x-coordinate (see #R63209, #R63304)",
            "Intermediate destination x-coordinate (see #R62926, #R62976)",
            "Intermediate destination y-coordinate (see #R62892, #R62976)",
            "Unknown (see #R25118, #R25307, #R62297)",
        )))
        char_buf_descs.append(("17", (
            "Unknown (see #R25118, #R25307, #R62297, #R63209)",
        )))
        char_buf_descs.append(("18,19", (
            "Uninterruptible subcommand routine address (see #R61948)",
        )))
        char_buf_descs.append(("20", (
            "Telephone dialling sound effect counter (see #R30675)",
            "Distance remaining for the hook to fly (see #R31713)",
            "Knockout and recovery delay counter (see #R64005)",
            "Unknown (see #R64338)",
        )))
        char_buf_descs.append(("21", (
            "Pre-knockout x-coordinate (see #R64005)",
            "Landing y-coordinate (see #R64338)",
        )))
        char_buf_descs.append(("22", (
            "Offset of the point reached in the current command list (relative to the start) (see #R61791)",
        )))
        char_buf_descs.append(("23,24", (
            "Start address of the current command list",
        )))
        char_buf_descs.append(("25,26", (
            "Start address of the primary command list",
        )))
        char_buf_descs.append(("27,28", (
            "Continual subcommand routine address (see #R61948)",
        )))
        char_buf_descs.append(("29", (
            "Unknown (see #R29100, #R29218, #R29559)",
        )))
        char_buf_descs.append(("30", (
            "Unknown (see #R29100, #R29218, #R29559)",
        )))
        char_buf_descs.append(("31", (
            "Unknown (see #R29100, #R29218, #R29559)",
        )))
        return char_buf_descs

    def _get_keypress_routines(self):
        keypress_routines = {}
        keypress_routines[24951] = 'Joystick'
        keypress_routines[25784] = 'Forward roll'
        keypress_routines[25789] = 'Somersault'
        keypress_routines[28635] = 'Toggle message speed'
        keypress_routines[29866] = 'Fuse'
        keypress_routines[30070] = 'Light on/off'
        keypress_routines[30092] = 'Pull/draw blind'
        keypress_routines[30458] = 'Telephone'
        keypress_routines[31204] = 'Get object'
        keypress_routines[31713] = 'Use hook'
        keypress_routines[60626] = 'Up'
        keypress_routines[60631] = 'Down'
        keypress_routines[60636] = 'Right'
        keypress_routines[60641] = 'Left'
        keypress_routines[62192] = 'Knock on or open a door'
        keypress_routines[64218] = 'Change disguise'
        return keypress_routines

    def _get_as_descs(self):
        as_descs = {}
        as_descs[0] = 'Sam standing/walking (phase 1)'
        as_descs[1] = 'Sam standing/walking (phase 2)'
        as_descs[2] = 'Sam standing/walking (phase 3)'
        as_descs[3] = 'Sam standing/walking (phase 4)'
        as_descs[4] = 'Sam with his arm up'
        as_descs[5] = 'Unused'
        as_descs[6] = 'Sam bending his knees, or rolling/somersaulting (phase 1)'
        as_descs[7] = 'Unused'
        as_descs[8] = 'Unused'
        as_descs[9] = 'Sam rolling (phase 3) or somersaulting (phase 4)'
        as_descs[10] = 'Sam rolling (phase 5) somersaulting (phase 6)'
        as_descs[11] = 'Unused'
        as_descs[12] = 'Unused'
        as_descs[13] = 'Unused'
        as_descs[14] = 'Unused'
        as_descs[15] = 'Sam rolling (phase 2) or somersaulting (phase 3)'
        as_descs[16] = 'Policeman standing/walking (phase 1)'
        as_descs[17] = 'Policeman standing/walking (phase 2)'
        as_descs[18] = 'Policeman standing/walking (phase 3)'
        as_descs[19] = 'Policeman standing/walking (phase 4)'
        as_descs[20] = 'Policeman with his arm up'
        as_descs[21] = 'Unused'
        as_descs[22] = 'Unused'
        as_descs[23] = 'Unused'
        as_descs[24] = 'Policeman standing/walking while carrying Sam (phase 1)'
        as_descs[25] = 'Policeman standing/walking while carrying Sam (phase 2)'
        as_descs[26] = 'Policeman standing/walking while carrying Sam (phase 3)'
        as_descs[27] = 'Policeman standing/walking while carrying Sam (phase 4)'
        as_descs[28] = 'Unused'
        as_descs[29] = 'Unused'
        as_descs[30] = 'Unused'
        as_descs[31] = 'Sam rolling (phase 4) or somersaulting (phase 5)'
        as_descs[32] = 'Gangster standing/walking (phase 1)'
        as_descs[33] = 'Gangster standing/walking (phase 2)'
        as_descs[34] = 'Gangster standing/walking (phase 3)'
        as_descs[35] = 'Gangster standing/walking (phase 4)'
        as_descs[36] = 'Gangster with his arm up'
        as_descs[37] = 'Unused'
        as_descs[38] = 'Unused'
        as_descs[39] = 'Unused'
        as_descs[40] = 'Gangster standing/walking while carrying Sam (phase 1)'
        as_descs[41] = 'Gangster standing/walking while carrying Sam (phase 2)'
        as_descs[42] = 'Gangster standing/walking while carrying Sam (phase 3)'
        as_descs[43] = 'Gangster standing/walking while carrying Sam (phase 4)'
        as_descs[44] = 'Unused'
        as_descs[45] = 'Unused'
        as_descs[46] = 'Unused'
        as_descs[47] = 'Unused'
        as_descs[48] = 'Lana standing/walking (phase 1)'
        as_descs[49] = 'Lana standing/walking (phase 2)'
        as_descs[50] = 'Lana standing/walking (phase 3)'
        as_descs[51] = 'Lana standing/walking (phase 4)'
        as_descs[52] = 'Lana with her arm up'
        as_descs[53] = 'Unused'
        as_descs[54] = 'Sniper (all phases)'
        as_descs[55] = 'Unused'
        as_descs[56] = 'Fat man standing/walking (phase 1)'
        as_descs[57] = 'Fat man standing/walking (phase 2)'
        as_descs[58] = 'Fat man standing/walking (phase 3)'
        as_descs[59] = 'Fat man standing/walking (phase 4)'
        as_descs[60] = 'Fat man with his arm up'
        as_descs[61] = 'Unused'
        as_descs[62] = 'Banknote (phase 1)'
        as_descs[63] = 'Unused'
        as_descs[64] = 'Skinny man standing/walking (phase 1) (unused)'
        as_descs[65] = 'Skinny man standing/walking (phase 2) (unused)'
        as_descs[66] = 'Skinny man standing/walking (phase 3) (unused)'
        as_descs[67] = 'Skinny man standing/walking (phase 4) (unused)'
        as_descs[68] = 'Skinny man with his arm up (unused)'
        as_descs[69] = 'Unused'
        as_descs[70] = 'Banknote (phase 2)'
        as_descs[71] = 'Unused'
        as_descs[72] = 'Daisy standing/walking (phase 1)'
        as_descs[73] = 'Daisy standing/walking (phase 2)'
        as_descs[74] = 'Daisy standing/walking (phase 3)'
        as_descs[75] = 'Daisy standing/walking (phase 4)'
        as_descs[76] = 'Daisy with her arm up'
        as_descs[77] = 'Unused'
        as_descs[78] = 'Unused'
        as_descs[79] = 'Unused'
        as_descs[80] = 'Lady in a hat standing/walking (phase 1)'
        as_descs[81] = 'Lady in a hat standing/walking (phase 2)'
        as_descs[82] = 'Lady in a hat standing/walking (phase 3)'
        as_descs[83] = 'Lady in a hat standing/walking (phase 4)'
        as_descs[84] = 'Lady in a hat with her arm up'
        as_descs[85] = 'Unused'
        as_descs[86] = 'Unused'
        as_descs[87] = 'Unused'
        as_descs[88] = 'Lady with big hair standing/walking (phase 1)'
        as_descs[89] = 'Lady with big hair standing/walking (phase 2)'
        as_descs[90] = 'Lady with big hair standing/walking (phase 3)'
        as_descs[91] = 'Lady with big hair standing/walking (phase 4)'
        as_descs[92] = 'Lady with big hair with her arm up'
        as_descs[93] = 'Unused'
        as_descs[94] = 'Unused'
        as_descs[95] = 'Unused'
        as_descs[96] = 'Girl with a pony tail standing/walking (phase 1)'
        as_descs[97] = 'Girl with a pony tail standing/walking (phase 2)'
        as_descs[98] = 'Girl with a pony tail standing/walking (phase 3)'
        as_descs[99] = 'Girl with a pony tail standing/walking (phase 4)'
        as_descs[100] = 'Girl with a pony tail with her arm up'
        as_descs[101] = 'Unused'
        as_descs[102] = 'Unused'
        as_descs[103] = 'Unused'
        as_descs[104] = 'Man in a bowler hat standing/walking (phase 1)'
        as_descs[105] = 'Man in a bowler hat standing/walking (phase 2)'
        as_descs[106] = 'Man in a bowler hat standing/walking (phase 3)'
        as_descs[107] = 'Man in a bowler hat standing/walking (phase 4)'
        as_descs[108] = 'Man in a bowler hat with his arm up'
        as_descs[109] = 'Unused'
        as_descs[110] = 'Unused'
        as_descs[111] = 'Unused'
        as_descs[112] = 'Man in a baseball cap standing/walking (phase 1)'
        as_descs[113] = 'Man in a baseball cap standing/walking (phase 2)'
        as_descs[114] = 'Man in a baseball cap standing/walking (phase 3)'
        as_descs[115] = 'Man in a baseball cap standing/walking (phase 4)'
        as_descs[116] = 'Man in a baseball cap with his arm up'
        as_descs[117] = 'Cat running (phase 1)'
        as_descs[118] = 'Cat running (phase 2)'
        as_descs[119] = 'Unused'
        as_descs[120] = 'Unused'
        as_descs[121] = 'Hook'
        as_descs[122] = 'Unused'
        as_descs[123] = 'Unused'
        as_descs[124] = 'Unused'
        as_descs[125] = 'Unused'
        as_descs[126] = 'Unused'
        as_descs[127] = 'Unused'
        return as_descs

    def _get_characters(self):
        characters = {}
        characters[215] = "Various"
        characters[216] = "Various"
        characters[217] = "Various"
        characters[218] = "Various"
        characters[219] = "Various"
        characters[220] = "Various"
        characters[221] = "Various"
        characters[222] = "Policeman; cat (mode 4)"
        characters[223] = "Policeman; cat (modes 0, 4)"
        characters[224] = "Gangster"
        characters[225] = "Fat Man (mode 2); Lana (mode 3); cat"
        characters[226] = "Daisy; cat (mode 4)"
        characters[227] = "Sniper; cat (mode 0)"
        characters[228] = "Banknote"
        characters[229] = "Banknote; hook"
        characters[230] = "Sam"
        return characters

    def get_skool_udg(self, y, x, show_chars=False):
        xh = x // 8
        yhp = 256 * (184 + y // 6)
        t = self.snapshot[xh + yhp]
        m = self.snapshot[21 + 256 * (160 + (x % 8))]
        ylp = 256 * (160 + 2 * (y % 6))
        if self.snapshot[t + 48896] & m:
            ylp += 3072
        r = self.snapshot[t + 128 + 256 * (184 + (x % 8))]
        z = self.snapshot[r + ylp]
        attr_ref = self.snapshot[32 + xh + yhp]
        attr = self.snapshot[attr_ref + 256 * (168 + (z & 3))]
        ref = self.snapshot[r + ylp + 256]
        if z & 128:
            # This location has a background and a foreground
            if z & 64:
                l = 128 + (z & 60) // 2
                if ref & 128:
                    l += 1
                a = ref | 128
            else:
                j = self.snapshot[64 + xh + yhp]
                k = self.snapshot[192 + xh + 65280]
                if self.snapshot[k + 32512]:
                    j |= 32
                if z & 4:
                    j = 64 * (j & 3) + (j & 32)
                a = ref | (j & 224)
                l = 159
            udg_addr = l + 32768
            data = self.snapshot[udg_addr:udg_addr + 2048:256]
            if show_chars:
                self._superimpose_sprite_udgs(data, x, y, 1)
            l = self.snapshot[a + 49152]
            h = 128
            for i in range(4):
                data[i] |= self.snapshot[l + 256 * h]
                data[i] &= self.snapshot[l + 256 * (h + 1)]
                h += 2
            l = self.snapshot[a + 49408]
            h = 128
            for i in range(4, 8):
                data[i] |= self.snapshot[l + 256 * h]
                data[i] &= self.snapshot[l + 256 * (h + 1)]
                h += 2
        else:
            udg_addr = ref + 256 * (128 + 2 * (z & 28))
            data = self.snapshot[udg_addr:udg_addr + 2048:256]
            if show_chars and z & 64 == 0:
                self._superimpose_sprite_udgs(data, x, y, 1)

        if show_chars:
            self._superimpose_sprite_udgs(data, x, y, 2)
            self._superimpose_sprite_udgs(data, x, y, 4)
        return Udg(attr, data)

    def _superimpose_sprite_udgs(self, skool_udg_data, x, y, z):
        initial_states = [self.snapshot[a:a + 4] for a in range(65198, 65408, 6)]
        initial_states += [self.snapshot[a:a + 4] for a in range(56870, 58918, 256)]
        sam_z = self.snapshot[58884]
        if sam_z == 0:
            sam_as, sam_x, sam_y, sam_z = 0, 229, 19, 1
        else:
            sam_as, sam_x, sam_y = self.snapshot[58880:58883]
        initial_states.append((sam_as, sam_x, sam_y, sam_z))
        for state, char_x, char_y, char_z in initial_states:
            if char_z == z:
                width, height = self._get_sprite_size(state)
                if char_y <= y < char_y + height and char_x <= x < char_x + width:
                    sprite_udg = self.get_sprite_udg(state, None, y - char_y, x - char_x)
                    for i in range(8):
                        skool_udg_data[i] = (skool_udg_data[i] | sprite_udg.data[i]) & sprite_udg.mask[i]

    def _get_sprite_size(self, state):
        width = 3
        height = 5
        if state & 127 >= 120:
            width = height = 2
        elif state % 8 == 7:
            width = 5
            height = 3
        return width, height

    def get_sprite_udg(self, state, attr, row, col, udg_page=None):
        width, height = self._get_sprite_size(state)
        as_norm = (state if state & 7 else state + 2) | 128
        udg_page = 199
        col = width - 1 - col if state & 128 else col
        ref_page = 215 + col * height + row
        udg = self._get_sprite_udg(as_norm, attr, ref_page, udg_page)
        if state & 128:
            self._flip_udg(udg)
        return udg

    def get_font_bitmap(self, character):
        return self.font.get(ord(character), (255, 255, 255, 255))

    def get_initial_states(self):
        return ()

class Udg:
    def __init__(self, attr, data, mask=None, attr_addr=None, q_addr=None, ref_addr=None, ref=None, udg_page=None, x=None, y=None):
        self.attr = attr
        self.data = data
        self.mask = mask
        self.attr_addr = attr_addr
        self.q_addr = q_addr
        self.ref_addr = ref_addr
        # We store the UDG reference now in case the snapshot changes before the
        # reference is looked up via self.snapshot[udg.ref_addr] (e.g. because a
        # saved snapshot is restored after alter_skool_udgs() has been called)
        self.ref = ref
        self.udg_page = udg_page
        self.x = x
        self.y = y
        # If ink or paper is set, it will override attr (see
        # skoolhtml.SkoolWriter.get_pixel_array())
        self.ink = None
        self.paper = None

def get_skool_class(name):
    if name == 'SkoolDaze':
        return SkoolDaze
    if name == 'BackToSkool':
        return BackToSkool
    if name == 'ContactSamCruise':
        return ContactSamCruise
    return SkoolGame

class TapInfo:
    def __init__(self, tap_address_table, max_tap):
        self.tap_address_table = tap_address_table
        self.min_tap = tap_address_table % 256
        self.max_tap = max_tap

    def get_tap_range(self):
        return range(self.min_tap, self.max_tap + 1, 2)
